/*****************************************************************************
 * This file is part of Kvazaar HEVC encoder.
 *
 * Copyright (c) 2021, Tampere University, ITU/ISO/IEC, project contributors
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * * Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 * 
 * * Redistributions in binary form must reproduce the above copyright notice, this
 *   list of conditions and the following disclaimer in the documentation and/or
 *   other materials provided with the distribution.
 * 
 * * Neither the name of the Tampere University or ITU/ISO/IEC nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 ****************************************************************************/

#include "greatest/greatest.h"

#include "test_strategies.h"

#include <string.h>

static coeff_t coeff_test_data[64 * 64];
static uint32_t expected_test_result;

static void setup()
{
  // Fill test data.
  coeff_t value = INT16_MIN;
  for (int i = 0; i < 64 * 64; i++) {
    coeff_test_data[i] = value;
    value += 16;
  }

  // Calculate expected result using the formula for an arithmetic sum.
  expected_test_result =
    2048 * (16 - INT16_MIN) / 2 +
    2048 * 2047 * 16 / 2;
}

TEST test_coeff_abs_sum()
{
  uint32_t sum = kvz_coeff_abs_sum(coeff_test_data, 64 * 64);
  ASSERT_EQ(sum, expected_test_result);
  PASS();
}

SUITE(coeff_sum_tests)
{
  setup();

  for (volatile int i = 0; i < strategies.count; ++i) {
    if (strcmp(strategies.strategies[i].type, "coeff_abs_sum") != 0) {
      continue;
    }

    kvz_coeff_abs_sum = strategies.strategies[i].fptr;
    RUN_TEST(test_coeff_abs_sum);
  }
}
